require 'amrita2/core'

module Amrita2
  class Template
    include Amrita2

    # Map amrita:type to element_option
    DefaultTypeMap = { 
      :use_args => { :use_args => true},
      :use_original_element=> { :use_original_element => true},
      :expand_by_member => { :expand_by_member => true },
      :erb => { :erb => true, :use_args=>true },
      :link => { :link_attr => [:href, [/javascript:.*/]] },
      :input => { :value_attr => :value },
      :checkbox => { :value_attr => :checked },
      :encoding_euc => { :encoding=>'EUC-JP' }
    }

    # keep id attribute of output if set.
    attr_accessor :keep_id
    # The name of attribute that will be used for template expandion by amrita.
    # You will need to set this if you use _id_ attribute fom DOM.
    # For expample, if this was set to "amrita_id", 
    # you can use amrita_id for amrita and id for DOM.
    attr_accessor :amrita_id

    attr_accessor :encoding, :spec, :ie_hack, :debug_source, :eval_cdata_as_erb     # :nodoc:    

    attr_accessor :type_map, :parts_id     # :nodoc:    

    attr_reader :mod, :element_option, :root     # :nodoc:    


    def initialize(&block)
      @keep_id = false
      @amrita_id = nil
      @setuped = false
      @spec = nil
      @spec_proc = block
      @type_map = DefaultTypeMap.clone
      @ie_hack = false
      @eval_cdata_as_erb = false
      @element_option = Hash.new
      @parts_template = Hash.new
      @parts_id = nil
      @encoding = $KCODE
      @debug_source = nil
    end

    def setup
      return if @setuped
      setup_document
      @spec = get_spec
      @mod = @spec.create_template_module(@rexml_doc, @rexml_doc)
      @expander = get_expander
      @setuped = true
    end

    def setup_document
      return if @root
      @rexml_doc = 
        if @eval_cdata_as_erb
          t = get_document_text
          t.gsub!(/<amrita:erb *id='(\w+)'>/, "<span id='\\1'><![CDATA[" )
          t.gsub!('</amrita:erb>', ']]></span>')
          REXML::Document.new(t)
        else
          get_document 
        end
      @root = @rexml_doc.root
    end

    def get_root_option
      ret = {
        :delete_id => (not @keep_id), 
        :key_attr_name=>( @amrita_id or :id),
        :ie_hack => @ie_hack,
        :eval_cdata_as_erb => @eval_cdata_as_erb,
        :debug_source => @debug_source,
      }
      case @encoding
      when "EUC" 
        ret[:encoding]='EUC-JP'
      when "SJIS" 
        ret[:encoding]='SHIFT_JIS'
      when "UTF8" 
        ret[:encoding]='UTF-8'
      when "NONE"
      else
        raise "unkonw encoding #{@encoding}"
      end
      ret
    end

    def get_spec
      if @spec 
        @spec
      elsif @spec_proc
        @spec = @spec_proc.call
      else
        @spec = make_default_spec(@root, @type_map, get_root_option, @element_option)
      end
      @spec = @spec.clone unless @spec.cloned
      @spec 
    end

    def get_expander
      @spec.generate_expander(@mod)
    end

    def use_erb(b=TOPLEVEL_BINDING)
      @eval_cdata_as_erb = b
      @simple_cdata = true
    end

    def define_parts(*parts_ids)
      setup_document
      parts_ids.each do |pid|
        case pid
        when Array
          pid.each { |id| define_parts(id) }
        when Symbol
          make_parts(@root, pid)
          raise "parts element #{pid} not found" unless @parts_template[pid]
        when Hash
          pid.each do |k, v|
            define_parts(k)
            @parts_template[k].define_parts(v) unless v == true
          end
        else
          raise "#{pid.class} not supported"
        end
      end
    end

    def make_parts(element, parts_id)
      element.each_element do |e|
        if e.attributes['id'] == parts_id.to_s
          e.parent.delete(e) 
          pt = @parts_template[parts_id] = TemplateElement.new(e)
          pt.element_option[parts_id]= element_option[parts_id]
          pt.parts_id = parts_id
          pt.debug_source = debug_source
          pt.use_erb(@eval_cdata_as_erb) if @eval_cdata_as_erb
        else
          make_parts(e, parts_id)
        end
      end
    end

    def [](x)
      case x
      when Symbol
        @parts_template[x]
      else
        x = { parts_id=>x} if parts_id
        expand(out="", x)
        Amrita2::SanitizedString[out]
      end
    end

    def expand(out, data=nil, &block)
      setup unless @setuped
      set_binding(@eval_cdata_as_erb) if @eval_cdata_as_erb.kind_of?(Binding)
      if data
        @expander.expand_template(out, data)
      else
        @mod.expand_template(out, &block)
      end
    end

    def prettyprint=(f)
      warn 'prettyprint is not supported'
    end

    def compact_space=(f)
      warn 'compact_space is not supported'
    end

    def escaped_id=(f)
      warn 'escaped_id is not supported'
    end

    def xml=(f)
      warn 'xml option is not supported'
    end

    def asxml=(f)
      warn 'asxml option is not supported'
    end

    def pre_format=(f)
      warn 'pre_format option is not supported'
    end

    def use_compiler=(f)
      warn 'use_compiler option is not supported'
    end

    def get_document
      nil
    end

   def make_default_spec(root, type_map, option, element_option)
      s = Core::RootSpec.new(option.clone)
      make_default_spec1(root, option, [s], type_map, element_option)
      s
   end

   def make_default_spec1(e, option, spec_stack, type_map, element_option)
     key_attr = option[:key_attr_name].to_s
     id_ = e.attributes[key_attr]
     if id_
       sym = id_.intern
       opt = get_option(e, option, type_map, element_option, sym)
       s = Core::DynamicElementSpec.new(sym, nil, opt)
       s.parent = spec_stack.first
       spec_stack.first.children << s
       spec_stack.unshift(s)
       e.elements.each do |c|
         make_default_spec1(c, option, spec_stack, type_map, element_option)
       end
       spec_stack.shift
       spec_stack.first.check_double_id if option[:check_double_id]
     else
       e.elements.each do |c|
         make_default_spec1(c, option, spec_stack, type_map, element_option)
       end
     end
   end

   def get_option(e, option, type_map, element_option, sym)
     ret = option.clone
     a = e.attributes
     typ = (a['amrita:type'] || element_option[sym])
     case typ
     when Symbol, String
       default_opt = type_map[typ.intern] || { typ.intern=>true } 
       ret.merge!(default_opt) if default_opt
       a.delete('amrita:type')
     when Hash
       default_opt = typ
       ret.merge!(default_opt) if default_opt
       a.delete('amrita:type')
     end
     if opt = a['amrita:option']
       ret.merge!(eval("{#{opt} }"))
       a.delete('amrita:option')
     end
     ret
   end
  end

  class TemplateFile < Template
    include Amrita2
    def initialize(fname, &block)
      super(&block)
      @fname = fname
    end

    def get_document
      REXML::Document.new(File.new(@fname))
    end

    def get_document_text
      File.new(@fname).read
    end
  end

  class TemplateText < Template
    include Amrita2
    def initialize(text, &block)
      super(&block)
      @text = text
    end

    def get_document
      REXML::Document.new(@text)
    end

    def get_document_text
      @text
    end
  end

  class TemplateElement < Template
    def initialize(e, &block)
      super(&block)
      @element = e 
    end

    def get_document
      doc = REXML::Document.new
      doc << @element
      doc
    end

    def get_document_text
      doc = REXML::Document.new
      doc << @element
      doc.to_s
    end
  end
end
