
require 'test/unit'
require 'test/unit/testsuite'
require 'test/unit/ui/console/testrunner'
require 'amrita2/template'
require 'rexml/document'
require 'open3'

# ruby -I~/amrita2_core/lib ~/amrita2_core/test/test_impl.rb   |ruby -pe '$_.sub!(/^ */, "") ' 


module Amrita2Test
  include Amrita2
  include Amrita2::Core

  class TestSanitize < Test::Unit::TestCase
    include Amrita2

    def test_sanitize
      t = "<>&"
      assert_equal("&lt;&gt;&amp;", t.amrita_sanitize)
      assert_equal("1", 1.amrita_sanitize)
    end

    def test_simple
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa)
      end
      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      m = spec.create_template_module(doc)

      out = ""
      m.expand_template(out) do |mm|
        mm.aaa('&<>')
      end

      assert_equal("<x id='aaa'>&amp;&lt;&gt;</x>", out)
      out = ""
      m.expand_template(out) do |mm|
        mm.aaa(SanitizedString['</x> INJECTION CODE <x>'])
      end

      assert_equal("<x id='aaa'></x> INJECTION CODE <x></x>", out)
    end


    def test_replace_attr
      spec = Amrita2::define_template_spec() do 
        dynamic_element(:aaa)
      end

      doc = REXML::Document.new  "<x id='aaa' class='c'>111</x>"
      m = spec.create_template_module(doc)

      out = ""
      m.expand_template(out) do |mm|
        mm.aaa('xxx', :class=>'<>"\'&')
      end
      assert_equal("<x class='&lt;&gt;&quot;&apos;&amp;' id='aaa'>xxx</x>", out)
    end
  end


  class TestCodeGenerator < Test::Unit::TestCase
    include Amrita2
    include Amrita2::Core

    def test_put_constant
      cg = CodeGenerator.new("out")
      cg.put_constant("aaa")
      cg.put_constant("bbb")
      cg.flush()
      assert_equal('out << "aaabbb"', cg.results.join)

      cg = CodeGenerator.new("out")
      cg.put_constant("aaa")
      cg.put_constant("bbb")
      assert_equal('out << "aaabbb"', cg.results.join)

      out = "111"
      eval cg.results.join
      assert_equal('111aaabbb', out)
    end

    def test_put_expression
      cg = CodeGenerator.new("out")
      cg.code("a = '11'")
      cg.put_constant("aaa")
      cg.put_expression("a * 2")
      cg.put_constant("bbb")
      out = ""
      eval cg.results.join(';')
      assert_equal('aaa1111bbb', out)
    end

    def test_if
      cg = CodeGenerator.new("out")
      cg.if_('a==b') do
        cg.put_constant("xyz")
      end
      assert_equal('if a==b;  out << "xyz";end', cg.results.join(';'))

      a=b=1
      out = ""
      eval cg.results.join("\n")
      assert_equal('xyz',out )

      a=2
      out = ""
      eval cg.results.join("\n")
      assert_equal('',out )
    end

    def test_if_else
      cg = CodeGenerator.new("out")
      cg.if_('a==b') do
        cg.put_constant("xyz")
        cg.else_ do
          cg.put_constant("abc")
        end
      end

      #put_constant cg.results.join("\n")
      assert_equal("if a==b;  out << \"xyz\";else;  out << \"abc\";end", cg.results.join(';'))

      a=b=1
      out = ""
      eval cg.results.join("\n")
      assert_equal('xyz',out )

      a=2
      out = ""
      eval cg.results.join("\n")
      assert_equal('abc',out )
    end

    def test_module
      cg = CodeGenerator.new("out")
      cg.define_module("M1") do
        cg.define_method("abc") do 
          cg.code("'abc'")
        end
      end

      m = Module.new
      m.module_eval cg.results.join("\n")
      assert_equal('abc',m::M1.abc)
    end

    def test_constant_def
      cg = CodeGenerator.new("out")
      cg.define_module("M1") do
        c1 = cg.define_constant(%[1])
        c2 = nil
        cg.define_method("abc") do 
          c2 = cg.define_constant(%[2])
          cg.code(c2)
        end
        cg.define_method("efg") do 
          c3 = cg.define_constant(%[1+2])
          cg.code("[#{c1}, #{c2}, #{c3}]")
        end
      end

      m = Module.new
      m.module_eval cg.results.join("\n")
      assert_equal(2,m::M1.abc)
      assert_equal([1, 2, 3],m::M1.efg)
    end

    def test_constant_def2
      cg = CodeGenerator.new("out")
      cg.define_module("M1") do
        c1 = cg.define_constant(%[1])
        c2 = nil
        cg.define_module("M2") do
          cg.define_method("abc") do 
            c4 = cg.define_constant(%[333])
            cg.code("[#{c4}]")
          end
        end
        cg.define_method("abc") do 
          c2 = cg.define_constant(%[2])
          cg.code(c2)
        end
        cg.define_method("efg") do 
          c3 = cg.define_constant(%[1+2])
          cg.code("[#{c1}, #{c2}, #{c3}]")
        end
      end

      m = Module.new
      m.module_eval cg.results.join("\n")
      assert_equal(2,m::M1.abc)
      assert_equal([1, 2, 3],m::M1.efg)
      assert_equal([333],m::M1::M2.abc)
    end


    def test_def_method_and_call
      cg = CodeGenerator.new("out")
      cg.define_module("M1") do
        meth_call = nil
        cg.define_method("abc") do 
          cg.code("a = 1; b = 2 ")
          meth_call = cg.define_method_and_call(%w(a b)) do
            cg.code("a + b")
          end
        end

        cg.define_method("efg") do 
          cg.code("a = 3; b = 4 ")
          cg.code(meth_call)
        end
      end
      m = Module.new
      m.module_eval cg.results.join("\n")
      assert_equal(3, m::M1::abc)
      assert_equal(7, m::M1::efg)
    end
  end

  class TestRuntime < Test::Unit::TestCase
    include Amrita2::Runtime

    def test_element
      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      e = doc.root
      assert_equal("x", e.name)
      assert_equal("id", e.attributes.keys.first)
      assert_equal("aaa", e.attributes["id"])
      assert_equal("111", e.text)
      assert_equal("<x id='aaa'>111</x>", e.to_s)

      e = new_element(:x, :id=>"aaa")
      assert_equal("x", e.name)
      assert_equal("id", e.attributes.keys.first)
      assert_equal("aaa", e.attributes["id"])
      assert_equal(nil, e.text)
      assert_equal("<x id='aaa'/>", e.to_s)

      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      e = doc.root
      assert_equal("<x id='aaa'>", start_tag(e))
      assert_equal("</x>", end_tag(e))

      e = new_element(:x, :id=>"aaa", :xxx=>"yy'")
      e.delete_attribute("id")
      assert_equal("<x xxx='yy&apos;'>", start_tag(e))

    end

    def test_mainstream
      out = ""
      out2 = ""
      new_context(out) do
        get_mainstream << "aaa"
        new_context(out2) do
          get_mainstream << "111"
          get_mainstream << "222"
        end
        get_mainstream << "bbb"
      end
      assert_equal("aaabbb", out)
      assert_equal("111222", out2)
    end

    def test_substream
      out = ""
      new_context(out) do
        get_mainstream << "aaa"
        new_context(get_substream(:xxx)) do
          get_mainstream << "111"
          get_mainstream << "222"
        end
        output_substream(:xxx)
        get_mainstream << "bbb"
      end
      assert_equal("aaa111222bbb", out)
    end

    def test_substream2
      out = ""
      new_context(out) do
        new_context(get_substream(:xxx)) do
          get_mainstream << "1"
          get_mainstream << "2"
        end
        new_context(get_substream(:yyy)) do
          get_mainstream << "3"
          get_mainstream << "4"
        end
        get_mainstream << "aaa"
        output_substream(:xxx)
        get_mainstream << "bbb"
        output_substream(:yyy)
        get_mainstream << "ccc"
      end
      assert_equal("aaa12bbb34ccc", out)
    end

    def test_binding
      o = Object.new
      def o.b1
        aaa = 'aaa'
        bbb = 'bbb'
        binding
      end
      def o.b2
        aaa = 'xxx'
        bbb = 'yyy'
        binding
      end

      out = ""
      new_context(out) do
        new_binding(o.b1) do
          new_context(get_substream(:xxx)) do
            get_mainstream << eval("aaa", get_binding)
            get_mainstream << eval("bbb", get_binding)
          end
        end
        new_context(get_substream(:yyy)) do
          set_binding(o.b2)
          get_mainstream << eval("aaa", get_binding)
          get_mainstream << eval("bbb", get_binding)
        end
        output_substream(:xxx)
        output_substream(:yyy)
      end
      assert_equal("aaabbbxxxyyy", out)
    end

    def test_context_data
      out = ""
      new_context(out, :data=>"aaa") do
        get_mainstream << get_context_data
        new_context_data("bbb") do
          get_mainstream << get_context_data
        end
        new_context(get_substream(:xxx)) do
          get_mainstream << get_context_data
          set_context_data("ccc")
          get_mainstream << get_context_data
        end
        get_mainstream << get_context_data
        output_substream(:xxx)
      end
      assert_equal("aaabbbaaaaaaccc", out)
    end
  end

  class TestDynamicElement < Test::Unit::TestCase
    include Amrita2::Core
    TEMPLATE_TEXT = <<-END
    <html>
    <body>
      <h1 id="title">title</h1>
      <p id="text">text</p>
      <table>
        <thead>
          <tr id="h"><th></th><th></th></tr>
        </thead>
        <tbody>
          <span></span>
          <tr id="d"><td id="aaa"></td><td id="bbb"></td></tr>
          <span></span>
        </tbody>
      </table>
      <div id="footer">powerd by <a href="http://www.ruby-lang.org/">ruby</a></div>
    </body>
    </html>
    END

    
    def test_dynamic_element
      spec = Amrita2::define_template_spec do 
        dynamic_element(:title)
      end
      assert(spec.kind_of?(RootSpec))
      assert_equal(:title, spec.children[0].sym)
      assert(spec.children[0].kind_of?(DynamicElementSpec))
      assert_equal(spec, spec.children[0].parent)

      doc = REXML::Document.new TEMPLATE_TEXT
      t = spec.create_template_module(doc)

      root = doc.root
      e = root.elements["body/h1"]
      assert(root.elements["body/h1"].amrita_spec)
      assert(:title, root.elements["body/h1"].amrita_spec.sym)
      assert(root.elements["body/h1"].amrita_spec.parent)
      assert(root.elements["body/h1"].amrita_spec.parent.kind_of?(RootSpec))
    end

    def test_dynamic_element2
      spec = Amrita2::define_template_spec do |root|
        dynamic_element(:title)
        root.dynamic_element(:text)
      end
      doc = REXML::Document.new TEMPLATE_TEXT
      t = spec.create_template_module(doc)

      root = doc.root
      e = root.elements["body/h1"]
      assert(root.elements["body/h1"].amrita_spec)
      assert(:title, root.elements["body/h1"].amrita_spec.sym)
      assert(root.elements["body/p"].amrita_spec)
      assert(:title, root.elements["body/p"].amrita_spec.sym)
    end
  end


  class TestAttr < Test::Unit::TestCase
    include Amrita2

    def test_replace_attr
      spec = Amrita2::define_template_spec() do 
        dynamic_element(:aaa)
      end

      doc = REXML::Document.new  "<x id='aaa' class='c'>111</x>"
      t = spec.create_template_module(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('xxx')
      end
      assert_equal("<x class='c' id='aaa'>xxx</x>", out)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('xxx', :class=>'cc')
      end
      assert_equal("<x class='cc' id='aaa'>xxx</x>", out)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('xxx', :xxx=>'yyy')
      end

      assert("<x class='c' xxx='yyy' id='aaa'>xxx</x>" ==  out ||
               "<x xxx='yyy' class='c' id='aaa'>xxx</x>" ==  out 
             )


      out = ""
      t.expand_template(out) do |m|
        m.aaa('xxx', :id=>nil, :xxx=>'yyy')
      end
      assert("<x class='c' xxx='yyy'>xxx</x>" == out ||
             "<x xxx='yyy' class='c'>xxx</x>" == out 
             )
    end

    def test_replace_attr_nest
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa) do 
          dynamic_element(:bbb)
        end
      end
      doc = REXML::Document.new "<a id='aaa' class='abc'><b id='bbb'>xxx</b></a>"
      t = spec.create_template_module(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa(:class=>'xyz') do |m2|
          m2.bbb('yyy')
        end
      end

      assert_equal("<a class='xyz' id='aaa'><b id='bbb'>yyy</b></a>", out)
    end

    def test_dash
      spec = Amrita2::define_template_spec do 
        dynamic_element(:meta)
      end

      doc = REXML::Document.new  "<meta id='meta' http-equiv='refresh' content='c' />"
      t = spec.create_template_module(doc)

      out = ""

      t.expand_template(out) do |m|
        m.meta(nil, :content=>10)
      end
      assert_equal("<meta id='meta' content='10' http-equiv='refresh'></meta>", out)
    end

    def test_multivalue
      mv1 = MultiValue.new(10, 20)
      assert_equal(10, mv1.value)
      assert_equal(20, mv1.value)
      assert_equal(10, mv1.value)
      mv2 = MultiValue.new('a', 'b', 'c')
      assert_equal('a', mv2.value)
      assert_equal(20, mv1.value)
      assert_equal('b', mv2.value)
      assert_equal('c', mv2.value)
      assert_equal(10, mv1.value)
    end

    def test_multivalue2
      t = TemplateText.new "<span id='aaa'></span>" 
      t.element_option[:aaa] = { :mv_attr => { :class=>%w(aaa bbb) } }

      t.expand(out="",
               :aaa => 'xxxx'
               )
      assert_equal("<span class='aaa'>xxxx</span>",
                   out)

      t.expand(out="",
               :aaa => (0..4)
               )
      assert_equal("<span class='aaa'>0</span><span class='bbb'>1</span><span class='aaa'>2</span><span class='bbb'>3</span><span class='aaa'>4</span>",
                   out)

      t.expand(out="",
               :aaa => (0..3)
               )
      assert_equal("<span class='aaa'>0</span><span class='bbb'>1</span><span class='aaa'>2</span><span class='bbb'>3</span>",
                   out)

      t.expand(out="",
               :aaa => (0..2)
               )
      assert_equal("<span class='aaa'>0</span><span class='bbb'>1</span><span class='aaa'>2</span>",
                   out)
    end

    def test_multivalue3
      t = TemplateText.new "<a id='aaa' href='x'></a>" 
      t.element_option[:aaa] = { :mv_attr => { :class=>%w(aaa bbb) } }
      #t.debug_source =STDOUT
      t.expand(out="",
               :aaa => (0..4)
               )
      assert_equal("<a href='x' class='aaa'>0</a><a href='x' class='bbb'>1</a><a href='x' class='aaa'>2</a><a href='x' class='bbb'>3</a><a href='x' class='aaa'>4</a>",
                   out)
    end

    def test_multivalue4
      t = TemplateText.new <<-END
        <table>
          <tr><th>name</th><th>author</th></tr>
          <tr id='table1'>
            <td id='name'></td>
            <td id='author'></td>
          </tr>
       </table>
      END
      t.element_option[:table1] = { :mv_attr => { :class=>%w(odd even) } }
      lang = Struct.new(:name, :author, :website)
      data = {
       :table1=>[ 
         lang.new("Ruby", "matz"),
         lang.new("Perl", "Larry Wall"),
         lang.new("Python","Guido van Rossum")
        ]
      }

      t.expand(out="", data)
      assert_equal("<table>\n          <tr><th>name</th><th>author</th></tr>\n          <tr class='odd'>\n            <td>Ruby</td>\n            <td>matz</td>\n          </tr><tr class='even'>\n            <td>Perl</td>\n            <td>Larry Wall</td>\n          </tr><tr class='odd'>\n            <td>Python</td>\n            <td>Guido van Rossum</td>\n          </tr>\n       </table>\n",
                   out)
    end
  end


  class TestTemplate < Test::Unit::TestCase
    include Amrita2

    def test_template
      t = TemplateText.new "<a id='aaa'><b id='bbb'></b></a>" 
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb('xxx')
        end
      end
      assert_equal("<a><b>xxx</b></a>", out)
    end

    def test_spec
      text = "<a id='aaa'><b id='bbb'></b></a>" 
      t = TemplateText.new(text) 
      t.spec =Amrita2::define_template_spec do
        dynamic_element(:aaa) do 
          dynamic_element(:bbb, :use_original_element=>true)
        end
      end
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb do |e|
            e.attributes['class'] = 'using_spec'
            e.text = 'using spec'
            e
          end
        end
      end
      assert_equal("<a id='aaa'><b class='using_spec' id='bbb'>using spec</b></a>", out)
    end

    def test_spec2
      text = "<a id='aaa'><b id='bbb'></b></a>" 
      t = TemplateText.new(text)  do
        Amrita2::define_template_spec do
          dynamic_element(:aaa) do 
            dynamic_element(:bbb, :use_original_element=>true)
          end
        end
      end
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb do |e|
            e.attributes['class'] = 'using_spec2'
            e.text = 'using spec2'
            e
          end
        end
      end
      assert_equal("<a id='aaa'><b class='using_spec2' id='bbb'>using spec2</b></a>", out)
    end

    def test_spec3
      text = "<a id='aaa'><b id='bbb'></b></a>" 
      t = TemplateText.new(text)
      def t.get_spec
        Amrita2::define_template_spec do
          dynamic_element(:aaa) do 
            dynamic_element(:bbb, :use_original_element=>true)
          end
        end
      end
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb do |e|
            e.attributes['class'] = 'using_spec3'
            e.text = 'using spec3'
            e
          end
        end
      end
      assert_equal("<a id='aaa'><b class='using_spec3' id='bbb'>using spec3</b></a>", out)
    end

    def test_spec4
      text = "<x id='aaa'><y id='bbb'>1</y><y id='bbb'>2</y></x>"
      t = TemplateText.new(text) do
        Amrita2::define_template_spec do
          dynamic_element(:aaa) do
            dynamic_element(:bbb) 
            dummy_element(:bbb) # 222 
          end
        end
      end

      data = {"aaa"=>{:bbb=>["111", "222"]}}
      t.expand(out="", data)
      assert_equal("<x id='aaa'><y id='bbb'>111</y><y id='bbb'>222</y></x>", out)
    end

    class TestPO
      class Child
        def bbb
          'xxx'
        end
      end
      def aaa
        Child.new
      end

      def active(m)
        m.active do |m|
          m.bbb('xxx')
          m.bbb('yyy')
          m.bbb('zzz')
        end
      end

      def attr(m)
        m.attr(:aaa=>'bbb') do |m|
          m.bbb('xxx', :yyy=>'zzz')
        end
      end
    end

    def test_presentatio_object
      t = TemplateText.new "<a id='aaa'><b id='bbb'></b></a>" 
      t.expand(out="", TestPO.new) 
      assert_equal("<a><b>xxx</b></a>", out)

      t = TemplateText.new "<a id='active'><b id='bbb'></b></a>" 
      t.expand(out="", TestPO.new) 
      assert_equal("<a><b>xxx</b><b>yyy</b><b>zzz</b></a>", out)


      t = TemplateText.new "<x id='outer'><a id='active'><b id='bbb'></b></a></x>" 
      t.expand(out="", :outer=>TestPO.new) 
      assert_equal("<x><a><b>xxx</b><b>yyy</b><b>zzz</b></a></x>", out)

      t = TemplateText.new "<x id='outer'><a id='active'><b id='bbb'></b></a></x>" 
      t.expand(out="", :outer=>[TestPO.new, TestPO.new]) 
      assert_equal("<x><a><b>xxx</b><b>yyy</b><b>zzz</b></a></x><x><a><b>xxx</b><b>yyy</b><b>zzz</b></a></x>", out)

      t = TemplateText.new "<x id='outer'><a id='attr'><b id='bbb'></b></a></x>" 
      t.expand(out="", 'outer'=>TestPO.new) 
      assert_equal("<x><a aaa='bbb'><b yyy='zzz'>xxx</b></a></x>", out)
    end

    Lang = Struct.new(:name, :author, :website)
    class LangClass
      attr_reader :name, :author
      def initialize(n, a)
        @name, @author = n, a
      end
    end

    def test_template2
      tmpl = TemplateText.new <<-END
        <table>
          <tr><th>name</th><th>author</th></tr>
          <tr id='table1'>
            <td id='name'></td>
            <td id='author'></td>
          </tr>
       </table>
      END
      data = {
       :table1=>[ 
         Lang.new("Ruby", "matz"),
         LangClass.new("Perl", "Larry Wall"),
         Lang.new("Python","Guido van Rossum")
        ]
      }
      tmpl.expand(out="", data)
      formated_out = ""
      Open3.popen3("w3m -T text/html") do |i, o, e|
        i.puts out
        i.close
        formated_out = o.read
      end
      assert_equal(<<END, formated_out)
 name       author
Ruby   matz
Perl   Larry Wall
Python Guido van Rossum

END
    end

    def test_template_hash
      tmpl = TemplateText.new <<-END
        <table>
          <tr><th>name</th><th>author</th></tr>
          <tr id='table1'>
            <td id='name'></td>
            <td id='author'></td>
          </tr>
       </table>
      END
      data = {
       :table1=>[ 
         { :name=>"Ruby", :author=>'matz'},
         { :name=>"Perl", :author=>'Larry Wall'},
         { :name=>"Python", :author=>'Guido van Rossum'}
        ]
      }
      tmpl.expand(out="", data)
      formated_out = ""
      Open3.popen3("w3m -T text/html") do |i, o, e|
        i.puts out
        i.close
        formated_out = o.read
      end
      assert_equal(<<END, formated_out)
 name       author
Ruby   matz
Perl   Larry Wall
Python Guido van Rossum

END
    end

    class WebSite < Struct.new(:name, :url)
      def site(m)
        m.site(self.name, :href=>self.url)
      end
    end

    def test_template3
      tmpl = TemplateText.new <<-END
        <table>
          <tr><th>name</th><th>author</th><th>website</th></tr>
          <tr id='table1'>
            <td id='name'></td>
            <td id='author'></td>
            <td id='website'><a id='site'/></td>
          </tr>
       </table>
      END
      data = {
       "table1"=>[ 
         Lang.new("Ruby", "matz", WebSite.new('Ruby home page', 'http://www.ruby-lang.org')),
         Lang.new("Perl", "Larry Wall"),
         { 'name'=>"Python",'author'=>"Guido van Rossum" ,
           'website'=> WebSite.new('Python Programing Language', 'http://www.python.org/')},
        ]
      }
      tmpl.expand(out="", data)
      #print out
      formated_out = ""
      Open3.popen3("w3m -T text/html") do |i, o, e|
        i.puts out
        i.close
        formated_out = o.read
      end
      #print formated_out
      assert_equal(<<END, formated_out)
 name       author               website
Ruby   matz             Ruby home page
Perl   Larry Wall
Python Guido van Rossum Python Programing Language

END
    end

    def test_doctype
      t = TemplateText.new <<-END
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" lang="en">
<head>
<title id="head_title" />
</head>
</html>
      END
      t.expand(out="", :head_title=>'test')
      assert_equal(<<-END, out)
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html lang='en' xmlns='http://www.w3.org/1999/xhtml'>
<head>
<title>test</title>
</head>
</html>
END
     end
  end

  class TestOption < Test::Unit::TestCase
    include Amrita2

    def test_keep_id
      t = TemplateText.new "<a id='aaa'><b id='bbb'></b></a>" 
      t.keep_id = true
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb('xxx')
        end
      end
      assert_equal("<a id='aaa'><b id='bbb'>xxx</b></a>", out)
    end

    def test_amrita_id
      t = TemplateText.new "<a aid='aaa'><b aid='bbb'></b></a>" 
      t.amrita_id = :aid
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb('xxx')
        end
      end
      assert_equal("<a><b>xxx</b></a>", out)
    end

    def test_amrita_id2
      t = TemplateText.new "<a aid='aaa'><b aid='bbb'></b></a>" 
      t.amrita_id = :aid
      t.keep_id = true
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb('xxx')
        end
      end
      assert_equal("<a aid='aaa'><b aid='bbb'>xxx</b></a>", out)
    end

    def test_ie_hack
      t = TemplateText.new  '<p><img src="foo.jpeg" alt="Foo" /><span id="aaa" /><img src="bar.jpeg" alt="Boo" /></p>'
      t.ie_hack = true
      t.expand(out="", :aaa=>'aaa')
      assert_equal("<p><img src='foo.jpeg' alt='Foo' />aaa<img src='bar.jpeg' alt='Boo' /></p>", out)

      t = TemplateText.new  '<p><img src="foo.jpeg" alt="Foo" /><span id="aaa" /><img src="bar.jpeg" alt="Boo" /></p>'
      t.ie_hack = false
      t.expand(out="", :aaa=>'aaa')
      assert_equal("<p><img src='foo.jpeg' alt='Foo'/>aaa<img src='bar.jpeg' alt='Boo'/></p>", out)
    end
  end

  class TestInlineTDL < Test::Unit::TestCase
    include Amrita2
    include Amrita2::Runtime

    def assert_equal_xml(a, b)
      a = REXML::Document.new(a, :ignore_whitespace_nodes=>:all).to_s
      b = REXML::Document.new(b, :ignore_whitespace_nodes=>:all).to_s
      assert_equal(a, b)
    rescue 
      puts a
      puts 
      puts b
      raise
    end

    def test_use_args
      t = TemplateText.new "<span id='aaa' amrita:type='use_args'>$1/$2</span>" 
      t.expand(out="",:aaa=>Tuple['abc', 'efg'])
      assert_equal("<span>abc/efg</span>", out)
      t.expand(out="",:aaa=>[t('aaa', 'bbb'), t('xyz', '789')])
      assert_equal("<span>aaa/bbb</span><span>xyz/789</span>", out)
    end

    def test_use_args_empty
      t = TemplateText.new "<span id='aaa' amrita:type='use_args'>$1/$2</span>" 
      t.expand(out="",:aaa=>[])
      assert_equal("", out)
      t.expand(out="",:aaa=>[t('aaa', 'bbb'), t('xyz', '789'), []])
      assert_equal("<span>aaa/bbb</span><span>xyz/789</span>", out)
    end

    def test_use_element
      t = TemplateText.new "<span id='aaa' amrita:type='use_original_element'>abc</span>" 
      p = proc do |e| 
        e.text += '789'; 
        e.attributes['class'] = 'xxx'
        e 
      end
      t.expand(out="",:aaa=> p)
      assert_equal("<span class='xxx'>abc789</span>", out)
    end

    def test_expand_by_member
      t = TemplateText.new "<span id='aaa' amrita:type='expand_by_member' amrita:option=':delete_tag=>true'><span id='year' /></span>" 
      t.expand(out="",:aaa=>Time.at(0))
      assert_equal("1970", out)
    end

    def test_value_attr
      t = TemplateText.new %[<span id='aaa' amrita:option=':value_attr=>"class"'>xxx</span>]
      t.expand(out="",:aaa=>'aaaa')
      assert_equal("<span class='aaaa'>xxx</span>", out)
    end

    def test_input
      t = TemplateText.new %[<input id='aaa' amrita:type='input' type='text' name='aaa[xxx]'/>]
      t.expand(out="",:aaa=>'aaaa')
      assert_equal("<input name='aaa[xxx]' type='text' value='aaaa'/>", out)
    end

    def test_link
      t = TemplateText.new %[<a id='aaa' amrita:type='link' href='http://yyy/'>xxx</a>]
      t.expand(out="",:aaa=>%w(http://xxx/  wwww) )
      assert_equal("<a href='http://xxx/'>wwww</a>", out)
    end

  end
end

if __FILE__ == $0
  include Amrita2Test
  cname = ARGV.shift
  if cname
    method = ARGV.shift
    if method
      eval "Test::Unit::UI::Console::TestRunner.run(#{cname}.new(\"#{method}\"))"
    else
      cls = eval(cname)
      Test::Unit::UI::Console::TestRunner.run(cls)
    end
  else
    suite = Test::Unit::TestSuite.new
    suite << TestCodeGenerator.suite
    suite << TestSanitize.suite
    suite << TestAttr.suite
    suite << TestRuntime.suite
    suite << TestSanitize.suite
    suite << TestDynamicElement.suite
    suite << TestInlineTDL.suite
    suite << TestOption.suite
    suite << TestTemplate.suite
    Test::Unit::UI::Console::TestRunner.run(suite)
  end
end
