
require 'test/unit'
require 'test/unit/testsuite'
require 'test/unit/ui/console/testrunner'
require 'amrita2/template'
require 'rexml/document'

# ruby -I~/amrita2/lib ~/amrita2/test/test_ja.rb   |ruby -pe '$_.sub!(/^ */, "") ' 

module Amrita2JapaneseTest
  include Amrita2

  class TestJapanese < Test::Unit::TestCase
    include Amrita2

    def setup
      @old_kcode = $KCODE
      $KCODE='e'
    end

    def teardown 
      $KCODE = @old_kcode
    end

    def test_ja
      spec = Amrita2::define_template_spec() do 
        dynamic_element(:aaa, :encoding=>"euc-jp") do
          dynamic_element(:bbb)
        end
      end
      doc = REXML::Document.new  "<?xml version='1.0' encoding='EUC-JP'?><x id='aaa'><y id='bbb' /></x>"
      t = spec.create_template_module(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa do |ma|
          ma.bbb('a')
        end
      end
      out = Iconv::iconv("euc-jp", "utf-8", out)[0]
      assert_equal("<x id='aaa'><y id='bbb'>a</y></x>", out)

      t.expand_template(out="") do |m|
        m.aaa do |ma|
          s = 'ܸ'
          ma.bbb(s)
        end
      end
      out = Iconv::iconv("euc-jp", "utf-8", out)[0]
      assert_equal("<x id='aaa'><y id='bbb'>ܸ</y></x>", out)
      t.expand_template(out="") do |m|
        m.aaa do |ma|
          ma.bbb('ܸ', :atr=>"ܸ°")
        end
      end
      out = Iconv::iconv("euc-jp", "utf-8", out)[0]
      assert_equal("<x id='aaa'><y id='bbb' atr='ܸ°'>ܸ</y></x>", out)
    end

    def test_sjis
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :encoding=>"shift_jis") do
          dynamic_element(:bbb)
        end
      end
      sjis_txt = Iconv::iconv("shift_jis", "euc-jp", "<?xml version='1.0' encoding='SHIFT_JIS'?><x id='aaa'><y id='bbb' /></x>")[0]
      doc = REXML::Document.new  sjis_txt
      t = spec.create_template_module(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa do |ma|
          ma.bbb('a')
        end
      end
      out = Iconv::iconv("euc-jp", "utf-8", out)[0]
      assert_equal("<x id='aaa'><y id='bbb'>a</y></x>", out)

      t.expand_template(out="") do |m|
        m.aaa do |ma|
          s = 'ܸ'
          s = Iconv::iconv("shift_jis", "euc-jp", s)[0]
          ma.bbb(s)
        end
      end
      out = Iconv::iconv("euc-jp", "utf-8", out)[0]
      assert_equal("<x id='aaa'><y id='bbb'>ܸ</y></x>", out)
      t.expand_template(out="") do |m|
        m.aaa do |ma|
          s = 'ܸ'
          s = Iconv::iconv("shift_jis", "euc-jp", s)[0]
          a = 'ܸ°'
          a = Iconv::iconv("shift_jis", "euc-jp", a)[0]
          ma.bbb(s, :atr=>a)
        end
      end
      out = Iconv::iconv("euc-jp", "utf-8", out)[0]
      assert_equal("<x id='aaa'><y id='bbb' atr='ܸ°'>ܸ</y></x>", out)
    end
  end
  class TestJapaneseTemplate < Test::Unit::TestCase
    include Amrita2
    def setup
      @old_kcode = $KCODE
    end

    def teardown 
      $KCODE = @old_kcode
    end

    def to_sjis(euc_txt)
      Iconv::iconv("shift_jis", "euc-jp", euc_txt)[0]
    end

    def to_utf8(euc_txt)
      Iconv::iconv("utf-8", "euc-jp", euc_txt)[0]
    end

    def from_utf(utf_txt)
      Iconv::iconv("euc-jp", "utf-8", utf_txt)[0]
    end

    def test_euc
      $KCODE='EUC'
      t = TemplateText.new "<?xml version='1.0' encoding='EUC-JP'?><div><span id='aaa'></span></div>" 
      t.expand(out="", :aaa=>'ܸ') 
      assert_equal("<div>ܸ</div>", from_utf(out))

      $KCODE='SJIS'
      t = TemplateText.new "<?xml version='1.0' encoding='EUC-JP'?><div><span id='aaa'></span></div>" 
      t.encoding='EUC'
      t.expand(out="", :aaa=>'ܸ') 
      assert_equal("<div>ܸ</div>", from_utf(out))
    end

    def test_sjis
      $KCODE='SJIS'
      euc_txt = "<?xml version='1.0' encoding='SHIFT_JIS'?><div><span id='aaa'></span></div>" 
      sjis_txt = to_sjis(euc_txt)
      t = TemplateText.new sjis_txt
      t.expand(out="", :aaa=>to_sjis('ܸ')) 
      assert_equal("<div>ܸ</div>", from_utf(out))

      $KCODE='UTF8'
      t = TemplateText.new sjis_txt
      t.encoding='SJIS'
      t.expand(out="", :aaa=>to_sjis('ܸ')) 
      assert_equal("<div>ܸ</div>", from_utf(out))

      $KCODE='UTF8'
      t = TemplateText.new sjis_txt
      t.encoding='SJIS'
      t.expand(out="", :aaa=>to_sjis('ܸ')) 
      assert_equal("<div>ܸ</div>", from_utf(out))
    end

    def test_utf8
      $KCODE='UTF8'
      euc_txt = "<?xml version='1.0' encoding='UTF-8'?><div><span id='aaa'></span></div>" 
      utf8_txt = to_utf8(euc_txt)
      t = TemplateText.new utf8_txt
      t.expand(out="", :aaa=>to_utf8('ܸ')) 
      assert_equal("<div>ܸ</div>", from_utf(out))

      $KCODE='EUC'
      euc_txt = "<?xml version='1.0' encoding='UTF-8'?><div><span id='aaa'></span></div>" 
      utf8_txt = to_utf8(euc_txt)
      t = TemplateText.new utf8_txt
      t.encoding='UTF8'
      t.expand(out="", :aaa=>to_utf8('ܸ')) 
      assert_equal("<div>ܸ</div>", from_utf(out))
    end
  end
end

if __FILE__ == $0
  include Amrita2JapaneseTest
  cname = ARGV.shift
  if cname
    method = ARGV.shift
    if method
      eval "Test::Unit::UI::Console::TestRunner.run(#{cname}.new(\"#{method}\"))"
    else
      cls = eval(cname)
      Test::Unit::UI::Console::TestRunner.run(cls)
    end
  else
    suite = Test::Unit::TestSuite.new
    suite << TestJapaneseTemplate.suite
    suite << TestJapanese.suite
    Test::Unit::UI::Console::TestRunner.run(suite)
  end
end
